# @package .
#  e4clim: Energy for CLimate Integrated Model.
#
#  e4clim: Energy for CLimate Integrated Model.

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions
